package PureMessage::Config;

use strict;
our $VERSION = "6.407";

use PureMessage::Version qw($PREFIX);
use PureMessage::_sys();

require Exporter;
our @ISA = qw(Exporter ActiveState::Config);
our @EXPORT_OK = qw(%config expand_config_filename expand_filename expand_tmpl_filename);

my $pmx_conf  = "$PREFIX/etc/pmx.conf";
my $pmx_cache = "$PREFIX/etc/pmx.conf.cache";
my $config;

# Try the quick and not fully accurate cache test,
# call the validate_cache() function to do a complete
# correct test.
my $pmx_conf_mtime = (stat $pmx_conf)[9];
if (my $pmx_cache_mtime = (stat $pmx_cache)[9]) {
    if ($pmx_cache_mtime == $pmx_conf_mtime) {
	# cached copy should be ok
	$config = do $pmx_cache;
	$config->{_cached_from} = $pmx_cache if $config;
    }
}

unless ($config) {
    $config = PureMessage::Config->new->as_hash;
    write_cache($config);
}

our %config;
*config = $config;  # alias;

# Inject %config into the filter namespaces:
for my $filt (keys %{$config{filter}}) {
    no strict 'refs';
    my $config = "${filt}::Config::config";
    *{$config} = $config{filter}{$filt};
}

# Find the license file:
if ($^O !~ /MSWin/) {
    if (my $lic_home = (getpwuid $config{user})[7]) {
	our $lic_file = "$lic_home/.ActiveState/ActiveState.lic";
    }
}
else {
    our $lic_file = "";
}

sub expand_config_filename {
    my ($base) = @_;
    return "$config{prefix}/etc/$base.conf";
}

sub expand_filename {
    my ($f) = @_;
    # XXX This doesn't match what expand_tmpl_filename() does but
    # the latter has the more sane behavior--we'll change this at
    # some future date once we know people aren't relying on this.
    $f = "$config{prefix}/$f" unless $f =~ m,^[./],;
    $f;
}

sub expand_tmpl_filename {
    my ($file, $lang) = @_;

    # return early for absolute paths
    return $file if $file =~ m,^/,;
    
    # check to see if old file exists
    if (-e "$config{prefix}/etc/$file") {
	return "$config{prefix}/etc/$file";
    }
    # else check for language
    #   if defined and translations exist use those templates
    elsif ((defined($lang)) && ($lang ne '')) {
	return "$config{prefix}/etc/templates/$lang/$file";
    }
    # otherwise default to English
    else {
	return "$config{prefix}/etc/templates/en/$file";
    }
}

sub validate_cache {
    my $forced_clean = shift || 0;
    # The full (and expensive) cache test.
    if ($forced_clean) {
        unlink($pmx_cache);
        if($config->{_cached_from}) {
	    %config = PureMessage::Config->new->as_hash;
	    write_cache(\%config);
        }
    } 
    elsif( my $pmx_cache_mtime = (stat $pmx_cache)[9] ) {
	require ActiveState::Config::Depend;
	$pmx_conf_mtime = ActiveState::Config::Depend::config_mtime_deps($pmx_conf);
	if ($pmx_cache_mtime != $pmx_conf_mtime) {
	    # cache stale
	    unlink($pmx_cache);
	    if ($config->{_cached_from}) {
		%config = PureMessage::Config->new->as_hash;
		write_cache(\%config);
	    }
	}
    }
}

sub write_cache {
    my $config = shift;
    my $fname = "$pmx_cache.$$";
    # We care about the *current* value of use_threads, so use %config.
    $fname .= PureMessage::_sys::tid() if $config{use_threads};
    if (open(my $f, ">", $fname)) {
	print $f "# This file is autogenerated from $pmx_conf.  Do not edit!!\n";
	require Data::Dump;
	my $x = Data::Dump::dump($config);
	$x =~ s/^\s+//mg;  # shorten it
	print $f "$x\n";
	if (close($f) &&
	    utime($pmx_conf_mtime, $pmx_conf_mtime, $fname) &&
	    rename($fname, $pmx_cache))
	{
	    $config->{_cached_from} = $pmx_cache;
	}
	else {
	    unlink($fname);
	}
    }
}


sub new {
    require ActiveState::Config;
    my $cls = shift;
    my $self = $cls->SUPER::new(
	strict		=> 1,
	default_type	=> ['#atom nodups', '#atom', '#list'],
	backup_ext	=> '.bak',
    );
    require PureMessage::Config::Validator;
    PureMessage::Config::Validator->validate($self);
    $self->parse_file($pmx_conf, @_);
    $self;
}

sub save {
    my $o = shift;
    my $t = time;
    utime $t, $t, $config{pmx_conf};
    $o->SUPER::save(@_);
}

1;
