package PureMessage::Config::Validator;

use strict;
use PureMessage::Version qw($PREFIX);
use PureMessage::_sys();
use ActiveState::Handy qw(ceil);
use ActiveState::Config qw(:path);

my $pmx_conf = "$PREFIX/etc/pmx.conf";
my $pmx_d    = "$PREFIX/etc/pmx.d";
my (@defs, @typedefs, %prop);

sub _concurrency_limit {
    my $curlim = PureMessage::_sys::get_phys_mem() / 40e6; # 40MB per thread
    $curlim = 10 if $curlim < 10;
    return int($curlim);
}

sub d_preload_interp {
    my $node = shift;
    my $r = $node->parser->root;
    (my $curlim_p = $node->path) =~ s/preload_interpreters/concurrency_limit/;
    my $pre = PureMessage::_sys::available_cpus() * 5;
    my $cur = path_value($r, $curlim_p);
    $pre = $cur if $pre > $cur;
    return $pre;
}

sub d_max_threads {
    my $node = shift;
    my $r = $node->parser->root;
    (my $use_threads = $node->path) =~ s/max_threads$/use_threads/;
    return path_value($r, $use_threads) ? 20 : 1;
}

sub d_max_pooled_procs {
    my $node = shift;
    my $r = $node->parser->root;
    (my $curlim_p = $node->path) =~ s/max_pooled_procs/concurrency_limit/;
    (my $max_threads_p = $node->path) =~ s/max_pooled_procs/max_threads/;
    my $curlim = path_value($r, $curlim_p);
    my $nthreads = path_value($r, $max_threads_p);
    return ceil($curlim / $nthreads);
}

sub d_min_pooled_procs {
    my $node = shift;
    my $r = $node->parser->root;
    (my $preload_p = $node->path) =~ s/min_pooled_procs/preload_interpreters/;
    (my $max_threads_p = $node->path) =~ s/min_pooled_procs/max_threads/;
    my $preload = path_value($r, $preload_p);
    my $nthreads = path_value($r, $max_threads_p);
    return ceil($preload / $nthreads);
}

sub _uid {
    my $uid = (stat "$PREFIX/bin/pmx")[4];
    defined($uid) or die "Can't stat $PREFIX/bin/pmx: $!";
    $uid;
}

sub _gid {
    my $gid = (stat "$PREFIX/bin/pmx")[5];
    defined($gid) or die "Can't stat $PREFIX/bin/pmx: $!";
    $gid;
}

# default block size
sub DEFAULT_BLOCK_SIZE () {8192}

# reasonable minimum value
sub MIN_BLOCK_SIZE     () {512}

# reasonable maximum value
sub MAX_BLOCK_SIZE     () {2_097_152}

sub _block_size_ok {
    my $blksize = shift or return 0;

    # check bounds
    return 0 unless $blksize >= MIN_BLOCK_SIZE() and $blksize <= MAX_BLOCK_SIZE();
    
    # make sure it is a power of of 2 integer
    return 0 unless ($blksize & ~(-$blksize)) == 0;

    return 1;
}

sub d_block_size {
    my $node = shift;
    
    my $f;
    
    # get value from a file in the quarantine, preferably
    # since it could be on a different filesystem
    my $r = $node->parser->root;
    my $qdir = path_value($r, 'quarantine_dir');
    $f = "$qdir/db.conf" if -f "$qdir/db.conf";
    
    # get it from pmx.conf instead
    unless ($f) {
        $f = $pmx_conf;
    }

    my $blksize = (stat $f)[11];
    
    return _block_size_ok($blksize) ? $blksize : DEFAULT_BLOCK_SIZE();
}

# Need some type definitions:
@typedefs = (
    ['#file',			\&v_file,	\&f_file],
    ['#logfile',		\&v_logfile,	\&f_logfile],
    ['#port',			\&v_port,	undef],
    ['#user',			\&v_user,	\&f_user],
    ['#group',			\&v_grp,	\&f_grp],
    ['#check_pkg',		\&v_pkg,	undef],
    ['#filemode',		\&v_mode,	\&f_mode],
    ['#blksize',		\&v_blksize,	undef],
    ['#re',		        \&v_re,	        undef],
    ['#smtp',                   \&v_smtp,       undef],
    ['#uint',		        \&v_uint,       undef],
);

# Declare the top-level options:
@defs = (
    # OPTION				TYPE		DEFAULT
    ['pmx_conf',			'#readonly',	$pmx_conf],
    ['pmx_d',				'#readonly',	$pmx_d],
    ['prefix',				'#readonly',	$PREFIX],
    ['log_dir',				'#file',	'var/log'],
    ['run_dir',				'#file',	'var/run'],
    ['pmx_db',                          '#str',         undef],
    ['quarantine',			'#str',		undef],
    ['quarantine_type',                 '#str',         'pmdb'],
    ['quarantine_dir',			'#file',	'var/qdir'],
    ['pmx_prd_store',                   '#str',         'pmx_quarantine'],
    ['pmx_resources_store',             '#str',         'pmx_quarantine'],
    ['stdout',				'#logfile',	'pmx_out'],
    ['stderr',				'#logfile',	'pmx_err'],
    ['log_to',				'#logfile',	'pmx_log'],
    ['message_log',                     '#logfile',     undef],
    ['log_utc',				'#bool',        0],
    ['async_release',			'#bool',        1],
    ['debug_level',			'#int',		0],
    ['warn_limit',			'#int',		0],
    ['umask',				'#filemode',	007],
    ['buffer_size',			'#int',		1_000_000],
    ['block_size',                      '#blksize',     \&d_block_size],
    ['time_connection',			'#bool',	0],
    ['time_message',			'#bool',	0],
    ['clone',				'#bool',	0],
    ['timeout',				'#duration',	0],
    ['max_proto_data_size',		'#int',		65535],
    ['pooled_proc_idle_limit',		'#duration',	'5m'],
    ['pooled_proc_connect_limit',	'#int',		20000],
    ['concurrency_limit_action',	'#enum tempfail wait accept',
							'tempfail'],
	['message_exception_action',    '#enum quarantine discard reject keep',
	                                'quarantine'],

    # The next six options are worth explaining:
    #
    # If you specify none of them (leave them all as defaults), then the
    # concurrency_limit is based on memory, use_threads is based on your OS,
    # and preload_interpreters is based on the number of CPUs. max_threads is
    # 20 if use_threads is enabled, otherwise it is ignored. max_pooled_procs
    # is the concurrency_limit divided by max_threads, and min_pooled_procs is
    # 1 if use_threads is enabled, otherwise it is preload_interpreters.
    #
    # If you specify the concurrency_limit, 
    ['concurrency_limit',		'#int',		_concurrency_limit()],
    ['use_threads',			'#bool',	0],
    ['preload_interpreters',		'#int',		\&d_preload_interp],
    ['max_threads',			'#int',		\&d_max_threads],
    ['max_pooled_procs',		'#int',		\&d_max_pooled_procs],
    ['min_pooled_procs',		'#int',		\&d_min_pooled_procs],

    ['thread_stack_size',		'#int',		1024],
    ['listen_queue_size',		'#int',		0],

    ['reinit_limit',			'#int',		2048],
    ['args',				'#list #str',	[]],
    ['perl_args',			'#list #str',	[]],
    ['user',				'#user',	_uid()],
    ['group',				'#group',	_gid()],
    ['internal-hosts',			'#list #str',	[]],
    ['quarantine_secret',		'#str'],
    ['mail_sender',                     '#smtp'],
    ['admin',				'#str'],

    ['milter#*.enabled',		'#bool',	1],
    ['milter#*.port',			'#port',	'inet:3366@localhost'],
    # smtp_talk_port is not a #port type because that does dup detection
    # which is not useful or correct for this setting
    ['milter#*.smtp_talk_port',		'#str',		''],
    ['milter#*.filter',			'#list',	\&d_filter],
    ['filter#*',			'#check_pkg'],
    ['blocklist#*.refresh_interval',    '#duration',    '5m'],
    ['blocklist#*.port',                '#port',        'inet:4466@localhost'],
    ['blocklist#*.blocklist_log',       '#logfile',     'blocklist_log'], 
    ['blocklist#*.connection_timeout',  '#duration',    '1s'],
    ['blocklist#*.block_dynamic',       '#bool',        'no'],
    ['blocklist#*.block_helo',          '#bool',        'no'],
    ['blocklist#*.reject_message',      '#str',         undef],
    ['mlog-stats#*.destination',        '#str',         'https://feedback.sophos.com/pmx/tr/v2'], 
    # The 'store' options should have their validators defined some place else
    # but there is no way to tell AS::Config where to load validators from when
    # !including another config file.
    ['store#*.dsn',                     '#str',         undef],
    ['store#*.username',                '#str',         undef],
    ['store#*.password',                '#str',         undef],
   
    # Sets the default language for the installation
    ['language',                        '#str',         'en'],
    
    ['relay_use_header',                '#bool',         0],
    ['relay_from_header',               '#str',         'received'],
    ['relay_from_regex',                '#re',           undef], #'^from .*? \((.*?)\)'
    ['relay_from_index',                '#uint',         0],
    ['use_helo_as_rdns',                '#bool',         undef],
    ['avoid_message_splitting',         '#bool',         0],
    ['avoid_message_splitting',         '#bool',         0],
    ['preserve_msgid',                  '#bool',         0],
    ['preserve_msgid_hdr_change',       '#bool',         0],
    ['quarantine_expire.expire_time',          '#duration',  '14d'],
    ['quarantine_expire.exec',                 '#str',         ''],
    ['quarantine_expire.archive',              '#bool',         1],
    ['quarantine_expire.archive_dir',          '#str', "home/archive"],

    # MIME parser limits - by default they're set inside the message object,
    # and these are only provided to allow customers (or ETS) to lower or
    # raise them if needed.
    ['mime_max_total_parts',            '#int',         undef],
    ['mime_max_part_depth',             '#int',         undef],
    ['mime_max_part_headers',           '#int',         512],

    ['active_mta',                      '#enum postfix sendmail unknown',  'unknown'],
    
    ['queuerunner.workers',             '#int',         '4'],
    ['queuerunner.stability',           '#duration',    '0'],
    ['queuerunner.timeout',             '#duration',    '10m'],
    ['queuerunner.sleep_timeout',       '#duration',    '30s'],
    ['queuerunner.min-backoff-time',    '#duration',    '180s'],
    ['queuerunner.max-backoff-time',    '#duration',    '4000s'],
    ['queuerunner.max-queue-lifetime',  '#duration',    '5d'],
    ['queuerunner.quarantine_secret',   '#bool',        0],
    ['quarantine_preserve_env_from',    '#bool',        0],
    
    # This address is used for redirecting bounced routed messages for
    # 'pmx_route :nobounce'
    # The value should be a valid email address or alias.
    ['nobounce-addr',   '#str', 'pmx-nobounce'],

);

# Propagate these top-level config options to milter#*:
%prop = map { $_ => 1 } (qw(
    preload_interpreters concurrency_limit reinit_limit max_pooled_procs
    min_pooled_procs pooled_proc_connect_limit debug_level umask
    concurrency_limit_action user group log_dir quarantine quarantine_dir
    stdout stderr log_to message_log log_utc args perl_args time_connection
    time_message clone use_threads timeout pooled_proc_idle_limit
    thread_stack_size listen_queue_size quarantine_type pmx_prd_store
    relay_use_header relay_from_header relay_from_regex relay_from_index
    use_helo_as_rdns max_proto_data_size
));

sub validate {
    my ($pkg, $p) = @_;
    $pkg->define_core_validators($p);
    $pkg->add_core_validators($p);
    $pkg->include_sub_validators($p);
}

sub define_core_validators {
    my ($pkg, $p) = @_;
    for my $def (@typedefs) {
	my ($type, $v, $f) = @$def;
	$p->typedecl($type, $v, $f);
    }
}

sub add_core_validators {
    my ($pkg, $p) = @_;
    for my $def (@defs) {
	my ($path, $type, $default) = @$def;
	$p->typedecl($path, $type) if defined $type;
	$p->default($path, $default) if defined $default;

	# Propagate this config option to milter#*, too:
	if ($prop{$path}) {
	    my $path2 = "milter#*.$path";
	    $p->typedecl($path2, $type) if defined $type;

	    # Unless the value has been defined, this default will fetch the
	    # value from the top-level.
	    $p->default($path2, sub {
		require ActiveState::Config;
		return ActiveState::Config::path_value($p->root, $path);
	    });
	}
    }
}

sub include_sub_validators {
    my ($pkg, $p) = @_;
    (my $path = $pkg) =~ s,::,/,g;
    for my $inc (@INC) {
	my $d = "$inc/$path";
	next unless -d $d;
	opendir(my $dir, $d) or next;
	while (my $conf = readdir $dir) {
	    my $pm = "$d/$conf";
	    next unless -f $pm;
	    my $cls = "${pkg}::${conf}";
	    next unless $cls =~ s/\.pm$//;
	    eval "require $cls" or next;
	    $cls->validate($p);
	}
    }
}

sub v_blksize {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    unless (_block_size_ok($v)) {
        $_[3] = d_block_size($node);
        $_[5] = 1;
        warn "Illegal block size '$v', setting to '$_[3]'.\n";
    }
}

sub v_file {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    unless ($v =~ m,^[./],) {
	$_[3] = "$PREFIX/$v";
	$_[5] = 1;
    }
}
sub f_file {
    my ($node, $val) = @_;
    $val =~ s/^\Q$PREFIX\E\///;
    $val;
}

sub v_logfile {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    require ActiveState::Config;
    my $log_dir = ActiveState::Config::path_value($p->root, 'log_dir');
    unless ($v =~ m,^[./],) {
	$_[3] = "$log_dir/$v";
	$_[5] = 1;
    }
}
sub f_logfile {
    my ($node, $val) = @_;
    my $root = $node->parser->root;
    require ActiveState::Config;
    my $log_dir = ActiveState::Config::path_value($root, 'log_dir');
    $val =~ s/^\Q$log_dir\E\///;
    $val;
}

{
    my %seen;
    sub v_port {
	my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
	require ActiveState::Config;
	my @parts = ActiveState::Config::pathsplit($path);
	$parts[$#parts] = 'enabled';
	my $q = ActiveState::Config::path(@parts);
	my $enabled = ActiveState::Config::path_value($p->root, $q);
	# Die if the port was previously seen in another milter.
	die "Duplicate port '$v'"
	    if $enabled and defined $seen{$v} and $seen{$v} ne $path;
	$seen{$v} = $path;
    }
}

sub v_user {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    unless ($v =~ /^\d+$/) {
	my $name = $v;
	$v = getpwnam($name);
	die "Bad user name '$name'" unless defined $v;
	$_[3] = $v;
	$_[5] = 1;
    }
}
sub f_user {
    my ($node, $val) = @_;
    if (defined(my $name = getpwuid($val))) {
	$val = $name;
    }
    $val;
}

sub v_grp {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    unless ($v =~ /^\d+$/) {
	my $name = $v;
	$v = getgrnam($name);
	die "Bad group name '$name'" unless defined $v;
	$_[3] = $v;
	$_[5] = 1;
    }
}
sub f_grp {
    my ($node, $val) = @_;
    if (defined(my $name = getgrgid($val))) {
	$val = $name;
    }
    $val;
}

sub v_pkg {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    (my $name = $node->name) =~ s#^PureMessage::##;
    eval "require PureMessage::Config::Validator::$name";
    warn "Unknown filter $name, settings ignored\n" if $@;
}

sub v_mode {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    die "invalid file mode" unless $v =~ /^[0-7]+$/;
    $v = oct $v;
    $_[3] = $v;
    $_[5] = 1;
    1;
}
sub f_mode {
    my ($node, $val) = @_;
    sprintf("0%02o", $val);
}

sub d_filter {
    my $node = shift;
    my $path = $node->path;
    (my $val  = $path) =~ s{milter#([^\.#]*)\.filter}{$1}i;
    $val;
}

sub v_re {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    $v =~ s,^/(.*)/$,$1,;
    eval {
        $v = qr/$v/;
    };
    if ($@) {
        die ("Invalid regular expression: '$_[3]'");
        return 0;
    }
}

sub v_smtp {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    my $IPv4 = qr/(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/;
    my $IPv6 = qr/([0-9a-fA-F:]+:[0-9a-fA-F:]+)/;
    my $IP   = qr/($IPv4|$IPv6)/;

    if ($v =~ /^smtp:.+:\d+$/) {
        if ($v !~ /^smtp:\[$IP\]:\d+$/) {
            die "invalid smtp format"
               if ( 
                    $v =~ /^smtp:[\[]/ or
                    $v =~ /^smtp:.+[\]]/ or
                    $v =~ /^smtp:$IP:\d+$/
                );
        }
    } else {
        die "invalid smtp format";
    }
}

sub v_uint {
    my ($p, $path, $node, $v, $append, $ctx, $ts) = @_;
    die "'$v' is not an unsigned integer" unless $v =~ /^\d+$/ and $v >= 0;
}

1;
